/***************************************************************************
 *   Copyright (C) 2007 by Mateusz Dworak aka compbatant                   *
 *   compbatant@linux.pl                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "FileListImpl.hpp"
#include "FilesEngine.hpp"
#include "Sort.hpp"
#include "Status.hpp"
#include "Mime.hpp"

#include "Command.hpp"
#include "VfsFactory.hpp"
#include "vfs/VfsLocal.hpp"
#include "Exceptions.hpp"



#include <boost/lexical_cast.hpp>
using boost::lexical_cast;

#include<boost/tokenizer.hpp>
using namespace boost;



FileListImpl::FileListImpl(FileListIcons *fli, Panel *_panel)
: fileListSync(NULL),
  fileListIcons(fli),
  panel(_panel),
  settings(singleton<Settings>::instance()),
  acquired(false),
  stopAllThreads(false),
  threadsNumber(0),
  acquiredNum(0)
{
	dragAndDrop=false;
	ascending=true;
	sortNr=0;
	
	VfsFactory &vfsFactory=singleton<VfsFactory>::instance();
	string dir="";
	
	if(panel->getParam()=="")
	{
		vfsName=settings.getDefaultVfsName();
		dir=settings.getDefaultPath();
	}
	else if(panel->getParam()=="left")
	{
		vfsName=settings.getLeftVfsName();
		dir=settings.getLeftPath();
	}
	else if(panel->getParam()=="right")
	{
		vfsName=settings.getRightVfsName();
		dir=settings.getRightPath();
	}
	else
	{
		
		typedef boost::tokenizer<boost::char_separator<char> >  tokenizer;
  		boost::char_separator<char> sep(":");
  		tokenizer tokens(panel->getParam(), sep);
	
		for (tokenizer::iterator tokIter = tokens.begin();tokIter != tokens.end(); ++tokIter)
   		{
   			string param=*tokIter;
   			if(param.size()>4)
   			{
   				if(param.substr(0,4)=="vfs=")
   				{
   					vfsName=param.substr(4,param.size());   				
   				}
   			}
   			
   		   	if(param.size()>10)
   			{
   				if(param.substr(0,10)=="directory=")
   				{
   					dir=param.substr(10,param.size());
   				}
   			}
   		}
		
	
	}
	
	vfs=vfsFactory.getVfs(vfsName);	
	vfs->init(panel,fileListIcons);
	//vfsSync=new VfsSync(vfs);
	
	vfsType=settings.getVfsType(vfsName);
	
	if(vfsType=="Local")
	isLocalVfs=true;
	
	
	headers=settings.getHeaders(vfsName);
	//headers.insert(headers.begin(),"name");	
	
	vector<Header> vec;
	vector<int>intVec=settings.getHeadersWidth(vfsName);
	for(int i=0;i<headers.size();++i)
	{
		Header h;
		h.name=headers[i];
		h.width=intVec[i];
		vec.push_back(h);
	}
	
	
	fileListIcons->setHeaders(vec);
	
	openDirectory(path(FilesEngine::expandPath(dir)));
	

	textFieldCommands=settings.getTextFieldCommands();
	
	vector<string>commands;
	vector< string >::iterator iter;
	for(iter=textFieldCommands.begin();iter!=textFieldCommands.end();++iter)
	{
		commands.push_back(*iter);
	}
	
	fileListIcons->setTextFieldCommands(commands);
	textFieldCommandSwitch(0);
	
}


FileListImpl::~FileListImpl()
{

	//delete vfsSync;

	delete vfs;
	
	
	delete fileListSync;
	
	/*FIXME*/
	//waitForThreads();
	
}

void FileListImpl::waitForThreads()
{
	map<ThreadCommand*,ThreadCommand*>::iterator iter;	

	for(iter=threadCommands.begin();iter!=threadCommands.end();++iter)
	{		
		ThreadCommand *tc=iter->second;		
		tc->setCanceledExecution();		
	}
	
	for(iter=threadCommands.begin();iter!=threadCommands.end();++iter)
	{		
		ThreadCommand *tc=iter->second;			
		tc->join();				
		delete tc;		
	}
	
	threadCommands.clear();

fileListIcons->stopProcessing();
	
}




void FileListImpl::openDirectory(path p)
{
	
	newDirectory=p.normalize();	
	runCommand("OpenDirectory");		
}

void FileListImpl::setCurrentDirectory(const path & dir)
{	
	currentDirectory = dir;
	panel->setPath( currentDirectory.string() );
	pathHistory.push_back(currentDirectory.string());
	
	fileListIcons->setStatusInfo("free: "+FilesEngine::printSize(FilesEngine::getFreeDiskSize(dir))+" files: "+lexical_cast<string>(files.size()));
	
}


void FileListImpl::openFile(path p)
{

	
	MimeInfo mimeInfo=settings.getMime(mime.getFileMime(p,isLocalVfs));	
	
	string commandName=mimeInfo.getDefaultCommand();

	if(commandName=="")
	commandName=settings.getMime(mime.getParentMime(mimeInfo)).getDefaultCommand();
	
	if(commandName=="")
	commandName=settings.getMime("special/unknown").getDefaultCommand();
	
	if(commandName!="")
	runCommand(commandName);
}


void FileListImpl::openIcon( FileIcon fileIcon )
{

	File file=fileIcon.getFile();
	//path p=currentDirectory;
	path p=file.getPath();
	p.normalize();

	if(file.isDirectory())
	{
		openDirectory(p);
	}
	else
	{
		settings.addFilesHistory(file.getPath().string());
		
		
		if(file.isUserExecutable())
		{
			if(file.getMime()=="application/x-executable")
			{
				FilesEngine::runFile(p);
				return;
			}
		}
		
		openFile(p);
	}

}

vector< string > FileListImpl::getDirectories( void )
{

vector <string> directories;

	for( filesIter=files.begin();filesIter!=files.end() ;++filesIter )
	{
		File file=filesIter->getFile();
		if(file.isDirectory() && !file.hidden())
		directories.push_back(file.getName());
	}
return directories;
}

void FileListImpl::openParentDir( )
{
	//if(acquired)
	//return;

	if(currentDirectory.string()!="/")
	openDirectory(currentDirectory.branch_path());
}

void FileListImpl::runCommand( string name, string param )
{

	if(acquired)
	return;

	if(name=="")
	return;
	
	Command *cmd=singleton<CommandsManager>::instance().getCommand(name,this,param);
	if(cmd)
	{		
		try
		{	
			cmd->execute();
			if(!cmd->isThread())
			delete cmd;
		}
		catch(...)
		{
			/*FIXME QUESTION DIALOG DON'T WORK IF NOT THREAD*/
		}
		
	}
	else
	{
	/*FIXME*/
	}	
}

vector< string > FileListImpl::getToolbarCommands( string name)
{

	return filterContexCommands( settings.getToolbarCommands( name ));
}

vector< string > FileListImpl::filterContexCommands( vector< string > commands )
{

	vector<string> contexCommands;
	vector<string>::iterator iter;

	bool anyFileSelected=fileListIcons->isFileIconSelected();
	
	for(iter=commands.begin();iter!=commands.end();++iter)
	{
		CommandInfo commandInfo=settings.getCommand(*iter);
	
		if(commandInfo.getNeedSelectedFile() && !anyFileSelected )
		{
		
			continue;
		}
		

		contexCommands.push_back(*iter);
	}
	
	return contexCommands;
}

void FileListImpl::sort( int nr )
{

	if(nr!=-1)
	{
		if(nr==sortNr)
		{
		ascending=!ascending;
		}
		sortNr=nr;
	}	
	
	
	string sortBy=headers.at(sortNr);
	if(sortBy=="name")
	{
		if(ascending)
		std::sort(files.begin(), files.end(),SortNameAscending());
		else
		std::sort(files.begin(), files.end(),SortNameDescending());
	}
	else if(sortBy=="size")
	{
		if(ascending)
		std::sort(files.begin(), files.end(),SortSizeAscending());
		else
		std::sort(files.begin(), files.end(),SortSizeDescending());
	}
	else if(sortBy=="mtime")
	{
		if(ascending)
		std::sort(files.begin(), files.end(),SortMTimeAscending());
		else
		std::sort(files.begin(), files.end(),SortMTimeDescending());
	}
	else if(sortBy=="ctime")
	{
		if(ascending)
		std::sort(files.begin(), files.end(),SortCTimeAscending());
		else
		std::sort(files.begin(), files.end(),SortCTimeDescending());
	}
	else if(sortBy=="atime")
	{
		if(ascending)
		std::sort(files.begin(), files.end(),SortATimeAscending());
		else
		std::sort(files.begin(), files.end(),SortATimeDescending());
	}
	else
	{
		if(ascending)
		std::sort(files.begin(), files.end(),SortStringFieldAscending(sortNr-1));
		else
		std::sort(files.begin(), files.end(),SortStringFieldDescending(sortNr-1));
	}
	
		std::string metric ( "B" );
  
  
  
		
	/*TODO*/
	//string sortBy=settings.getSortHeader( vfsName );
	//settings.getHeaders(vfsName);
		
		
		
		//removeFileIcons();
		
		fileListIcons->clearItems();
		
		/*		
			vector<FileIcon>filesIcons;
					
			for(filesIter=files.begin();filesIter!=files.end();++filesIter)
			{
				//FileIcon fileIcon=createFileIcon(*filesIter);
				//files.push_back(fileIcon);
				//filesIcons.push_back(fileIcon);
				//filesMap[fileIcon.getFile().getPath().string()]=fileIcon;	
										
			}
		*/	
		filesMap.clear();
		int pos=0;
		for(filesIter=files.begin();filesIter!=files.end();++filesIter)
		{
			filesMap[filesIter->getFile().getPath().string()]=pos;
			++pos;
		}
		
		fileListIcons->addFileIcons(files);	
			
	
	
}


void FileListImpl::reloadDirectory()
{

	openDirectory(currentDirectory);
}



FileIcon FileListImpl::createFileIcon(File file)
{

	FileIcon fileIcon(file);

	MimeInfo mimeInfo=settings.getMime(file.getMime());
	MimeInfo parentMimeInfo=settings.getMime(mime.getParentMime(file.getMime()));
	MimeInfo unknownMime=settings.getMime("special/unknown");

		string icon=mimeInfo.getIcon();
		if(!icon.size())
		{
		icon=parentMimeInfo.getIcon();
			if(!icon.size())
			{		
			icon=unknownMime.getIcon();
			}
		}
		
		if(file.isCharacterDevice())
		icon="blockdevice";		
		else if(file.isCharacterDevice())
		icon="chardevice";
		else if(file.isSocket())
		icon="socket";
		
		fileIcon.setIconName( icon );
	
		string color=mimeInfo.getColor();
		string backColor=mimeInfo.getBackColor();
		
		if(color=="")
		{
			color=parentMimeInfo.getColor();
				if(color=="")
				{
					color=unknownMime.getColor();
				}
		}
	
		if(backColor=="")
		{
			backColor=parentMimeInfo.getBackColor();
				if(backColor=="")
				{
					backColor=unknownMime.getBackColor();
				}
		}
	
		fileIcon.setColor(color);
		fileIcon.setBackColor(backColor);
		
	if(file.isSymlink())
		fileIcon.setExtraIconName("symlink");	
	if(file.isExecutable() && !file.isDirectory())
		fileIcon.setExtraIconName("execute");		
		
		
		
	return fileIcon;	
	

}



void FileListImpl::textFieldExecute( string data )
{

	textFieldData=data;
	runCommand(textFieldCommandName,"execute");
}

void FileListImpl::textFieldCommandSwitch(int nr)
{
	textFieldCommandName=textFieldCommands[nr];
	runCommand(textFieldCommandName,"switch");
}

void FileListImpl::clicked( void )
{

	vector<FileIcon> selectedFiles=getFileListIcons()->getSelectedFileIcons( );
	int selectedNum=selectedFiles.size();
	if(selectedNum>0)
	{
		if(selectedNum>1)
		{
		fileListIcons->setStatusInfo( "files: "+lexical_cast<string>(files.size())+" selected: " + lexical_cast<string>(selectedNum));
		}
		else
		{
			FileIcon fileIcon=selectedFiles.at(0);
			fileListIcons->setStatusInfo(fileIcon.getFile().getName()+" "+fileIcon.getFile().getMime()+" "+FilesEngine::printSize(fileIcon.getFile().getSize()));
		}
	}
	else
	{
	fileListIcons->setStatusInfo("free: "+FilesEngine::printSize(FilesEngine::getFreeDiskSize(currentDirectory))+" files: "+lexical_cast<string>(files.size()));
	}
}

Vfs * FileListImpl::getVfs()
{
	//return vfsSync;
	return vfs;
}

void FileListImpl::stopThreadProcessing(ThreadCommand * tc)
{
	//fileListIcons->stopProcessing();

	//if(tc->running())
	//tc->cancel(); /*FIXME join*/
		
	//threadCommands.erase(tc);
	
	//if(threadCommands.size()==0)
	//fileListIcons->stopProcessing();

	--threadsNumber;
	if(threadsNumber==0)
	fileListIcons->stopProcessing();
}

void FileListImpl::startThreadProcessing(ThreadCommand * tc)
{
	fileListIcons->startProcessing();
	threadCommands[tc]=tc;
	++threadsNumber;
}


vector< string > FileListImpl::getAvailableCommandsForSelectedFiles()
{

		vector<FileIcon> files=fileListIcons->getSelectedFileIcons();		
		vector<FileIcon>::iterator iter;
		vector<string> commands;
					
			for(iter=files.begin();iter!=files.end();++iter)
			{

				MimeInfo mimeInfo=settings.getMime(iter->getFile().getMime());
				MimeInfo mimeInfo2=settings.getMime(Mime::getParentMime(iter->getFile().getMime()));
				
			
				vector<string> commandsName=mimeInfo.getCommands();
				vector<string> commandsName2=mimeInfo2.getCommands();
				
				commandsName.insert(commandsName.end(),commandsName2.begin(),commandsName2.end());
				if(files.size()>1)
				{
					if(mimeInfo.getDefaultCommand()!="")
					commandsName.push_back(mimeInfo.getDefaultCommand());
						if(mimeInfo.getDefaultCommand()!=mimeInfo2.getDefaultCommand())
						{
							if(mimeInfo2.getDefaultCommand()!="")
							commandsName.push_back(mimeInfo2.getDefaultCommand());
						}
				}
				
				vector<string>::iterator iter2;
				/*TODO tylko czesc wspolna ma byc*/
				if(iter==files.begin()) //adding
				{
					for(iter2=commandsName.begin();iter2!=commandsName.end();++iter2)
					{
						if(iter==files.begin())
						{
							if(find(commands.begin(),commands.end(),*iter2)==commands.end())
							commands.push_back(*iter2);
						}
									
					}
				}
				else //removing
				{
		
					for(iter2=commands.begin();iter2!=commands.end();++iter2)
					{			
						if(commandsName.size()==0 || find(commandsName.begin(),commandsName.end(),*iter2)==commands.end())
						{	
							if(iter2+1==commands.end())	
							{
							commands.erase(iter2);
							break;
							}
							else
							{	
							commands.erase(iter2);
							iter2=commands.begin();
							}
						}		
						
					}
				}
			}
			
	return commands;		
}

path FileListImpl::accessFile(path filePath)
{

	VfsLocal vfsLocal;
	
	srand((unsigned)time(0)); 
	path tmpfile;
	
	do
	{
		int randomInteger = rand(); 	
		string randomPart=lexical_cast<string>(randomInteger);	
		tmpfile=path("/tmp/nao")/path(randomPart+string("__")+string(filePath.leaf()));	
	}
	while(FilesEngine::exists(tmpfile));		
				
			
	File file=vfs->readFile(filePath);
	vfs->copyFromFile(file,tmpfile,NULL,NULL);

	File fileTmp=vfsLocal.readFile(tmpfile);
	FileModificationTime fmt;
	fmt.fileName=filePath;
	fmt.modificationTime=fileTmp.getModificationTime();
	tmpFiles[tmpfile]=fmt;
	
	return tmpfile;
}

void FileListImpl::releaseFile(path filePath, bool commit)
{
	VfsLocal vfsLocal;

	map<path,FileModificationTime>::iterator iter;
	
	iter=tmpFiles.find(filePath);

	if(iter!=tmpFiles.end())
	{
		if(!commit)
		{
			vfsLocal.removeFile(filePath);
		}
		else
		{
			File fileTmp=vfsLocal.readFile(filePath);
			
			if(fileTmp.getModificationTime()!=iter->second.modificationTime)
			{
			vfs->copyToFile(fileTmp,iter->second.fileName,NULL,NULL);
			}
		}
		
	vfsLocal.removeFile(filePath);	
		
	}

}

vector< path > FileListImpl::getSelectedFileIconsPaths(void )
{
	vector<string>::iterator iter;
	vector<path> selectedFiles;
	
		if(!dragAndDrop)
		{
			vector<FileIcon> files=getFileListIcons()->getSelectedFileIcons( );		
			vector<FileIcon>::iterator iter;
					
			for(iter=files.begin();iter!=files.end();++iter)
			{
				if(iter->getFile().getName()!="..")
				selectedFiles.push_back(iter->getFile().getPath());
			}
			
		}
		else
		{
			for(iter=droppedFileNames.begin();iter!=droppedFileNames.end();++iter)
			{
				selectedFiles.push_back(path(*iter));
			}	
		}
		
	return selectedFiles;	
}


vector< string > FileListImpl::getSelectedFileIcons(void )
{

	vector<string>::iterator iter;
	vector<string> selectedFiles;
	
		if(!dragAndDrop)
		{
			vector<FileIcon> files=getFileListIcons()->getSelectedFileIcons( );		
			vector<FileIcon>::iterator iter;
					
			for(iter=files.begin();iter!=files.end();++iter)
			{
				if(iter->getFile().getName()!="..")
				selectedFiles.push_back(iter->getFile().getPath().string());
			}
			
		return selectedFiles;	
		}
		else
		{
			return droppedFileNames;	
		}
	
}


void FileListImpl::removeFileIcons()
{
	filesMap.clear();
	files.clear();
	fileListIcons->clearItems();
}

void FileListImpl::removeFileIcon(string p)
{
	vector<FileIcon>::iterator filesIter;

	File file;
	path pp=p;
	file.setName(pp.leaf());
	file.setParentDirectory(pp.branch_path());
	FileIcon fileIcon(file);


	filesMapIter=filesMap.find(p);
	if(filesMapIter!=filesMap.end())
	{
		
		int pos=filesMapIter->second;	
		
		fileListIcons->removeFileIcon(pos);	
		files.erase(files.begin()+pos);		
		filesMap.clear();
		
		pos=0;
		for(filesIter=files.begin();filesIter!=files.end();++filesIter)
		{
			filesMap[filesIter->getFile().getPath().string()]=pos;
			++pos;
		}
		
	}
	else
	{
		/*TODO add exception?*/		
	}
/*	
	filesIter=find(files.begin(),files.end(),fileIcon);
	if(filesIter!=files.end())
	{
		
		fileListIcons->removeFileIcon(filesIter-files.begin());
		filesMapIter=filesMap.find(file.getPath().string());
			
		if(filesMapIter!=filesMap.end())
		filesMap.erase(filesMapIter);
			
		files.erase(filesIter);
			
	}
*/	
}

void FileListImpl::replaceFileIcon(FileIcon fileIcon)
{
	filesMapIter=filesMap.find(fileIcon.getFile().getPath().string());
	if(filesMapIter!=filesMap.end())
	{
		fileListIcons->replaceFileIcon(filesMapIter->second,fileIcon);
	}
	else
	{
	throw ExceptionFileIconNotFound();
	}
}

void FileListImpl::addFile(string p)
{
	filesMapIter=filesMap.find(p);
	if(filesMapIter==filesMap.end())
	{
		File file=vfs->readFile(p);
		file=vfs->readFullFile(file);
		addFileIcon(file);
	}
}


void FileListImpl::addFileIcon(File file)
{
	filesMapIter=filesMap.find(file.getPath().string());
	if(filesMapIter==filesMap.end())
	{
		
		FileIcon fileIcon=createFileIcon(file);
		
		files.push_back(fileIcon);
		filesMap[file.getPath().string()]=files.size()-1;
		
		
		fileListIcons->addFileIcon(fileIcon);
	}	

}





FileIcon FileListImpl::getFileIcon(string p)
{

	filesMapIter=filesMap.find(p);
	if(filesMapIter!=filesMap.end())
	return files[filesMap[p]];
	else
	{
	throw ExceptionFileIconNotFound();
	}
}

void FileListImpl::refreshFileIcon(string p)
{
	
	filesMapIter=filesMap.find(p);
	if(filesMapIter!=filesMap.end())
	{
		//return filesMap[p];
		File file=vfs->readFile(path(p));
		file=vfs->readFullFile(file);
		replaceFileIcon(createFileIcon(file));
	}
	else
	throw ExceptionFileIconNotFound();
}

vector< string > FileListImpl::getAllFilesPaths()
{
	vector<string>paths;
	for(filesIter=files.begin();filesIter!=files.end();++filesIter)
	{
		paths.push_back(filesIter->getFile().getPath().string());
	}
	return paths;
}

void FileListImpl::selectFileIcon(string p)
{
	File file;
	path pp=p;
	file.setName(pp.leaf());
	file.setParentDirectory(pp.branch_path());
	FileIcon fileIcon(file);

	filesMapIter=filesMap.find(fileIcon.getFile().getPath().string());
	if(filesMapIter!=filesMap.end())
	{
		fileListIcons->selectFileIcon(filesMapIter->second);
	}
	else
	{
	throw ExceptionFileIconNotFound();
	}
}

void FileListImpl::setFiles(vector< File > filesVec)
{
	files.clear();
	filesMap.clear();
	vector<File>::iterator iter;
		
	for(iter=filesVec.begin();iter!=filesVec.end();++iter)
	{
		FileIcon fileIcon=createFileIcon(*iter);
		//filesMap[iter->getPath().string()]=fileIcon;
		files.push_back(fileIcon);
	}
	
}

FileIcon FileListImpl::getCurrentFileIcon()
{
	return fileListIcons->getCurrentFileIcon();
}

void FileListImpl::setOpposite(FileList * theValue)
{
	opposite = theValue;

}

void FileListImpl::release()
{
	--acquiredNum;
}

void FileListImpl::acquire()
{
	++acquiredNum;
}

void FileListImpl::tick()
{
	map<ThreadCommand*,ThreadCommand*>::iterator iter;
	for(iter=threadCommands.begin();iter!=threadCommands.end();++iter)
	{		
		ThreadCommand *tc=iter->second;			
		if(!tc->running())
		{
			tc->join();				
			delete tc;
			threadCommands.erase(iter);
			break;
		}
			
	}
}


