/***************************************************************************
 *   Copyright (C) 2007 by Mateusz Dworak aka compbatant                   *
 *   compbatant@linux.pl                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "VfsFtp.hpp"
#include "FilesEngine.hpp"
#include "Exceptions.hpp"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>


REGISTER_FACTORABLE(VfsFtp);

class FtpOpenedDirectory 
{
	public:
	
	vector<File> files;
    	vector<File>::iterator iter;

};


VfsFtp::VfsFtp() : VfsBase("remote"), connected(false)
{
}


VfsFtp::~VfsFtp()
{
	if(connected && ftp.isConnected())
	ftp.Quit();
}


void VfsFtp::connect()
{
	string user=params["user"];
	if(user=="")
	user=FilesEngine::getCurrentUser();
	
	string password=params["password"];
	string hostname=params["hostname"];
	string port=params["port"];
	if(port=="")
	port="21";
	
	
	connectionDescription=user+"@"+hostname;
	
	if(ftp.Connect(string(hostname+":"+port).c_str())==0)
	{
		throw ExceptionConnectionRefused(connectionDescription,"");
	}
	
	#ifdef HAVE_SSL
	ftp.NegotiateEncryption();
	#endif
	
	if(ftp.Login(user.c_str(),password.c_str())==0)
	{
		throw ExceptionAuthenticationFailed(connectionDescription,"");
	}
	
	connected=true;
}

Directory VfsFtp::openDir(path & p, Cancelable * cancelable)
{	
	path pp=p;
	
	if(!connected)
	{
		connect();	
		char initialDir[256];
		ftp.Pwd(initialDir,256);
		pp=path(string(initialDir));
		
	}
	
	if(!ftp.isConnected())
	throw ExceptionConnectionLost(connectionDescription,"");
	
	p=pp;
	
	stringstream streambuffer;	
	
	ftp.Dir(streambuffer,pp.string().c_str());
	string field;
	
	char line[500];
	
	FtpOpenedDirectory *ftpDir=new FtpOpenedDirectory();
	
	while(!streambuffer.eof())
	{
		
		streambuffer.getline(line,500);
		
		stringstream linebuffer(line);
		File file;
		
		linebuffer >> field; //atrib

		file.setMode(FilesEngine::stringToPremissions(field));
		
		if(field.at(0)=='d')
		file.setDirectory();		
		
		linebuffer >> field; //
		linebuffer >> field; //user
		file.setUser(field);
		linebuffer >> field; //group
		file.setGroup(field);
		linebuffer >> field; //size
		file.setSize(atoi(field.c_str()));
		linebuffer >> field; //date
		linebuffer >> field; //date
		linebuffer >> field; //date
		
		string name;
		string n;
		
			while ( linebuffer >> n )
			{
				if(name=="")
				name=n;
				else
				name=name+" "+n;
			}
			
			if(name!="")
			{
			file.setName(name);
			file.setParentDirectory(pp);
			ftpDir->files.push_back(file);
			}
		}
	
	

	ftpDir->iter=ftpDir->files.begin();
	
	Directory directory(this,ftpDir,pp);
	return directory;
}

void VfsFtp::closeDir(Directory * directory)
{
	FtpOpenedDirectory *ftpDir=static_cast<FtpOpenedDirectory *>(directory->getData());
	delete ftpDir;
}

File VfsFtp::readDir(Directory * directory)
{

	FtpOpenedDirectory *ftpDir=static_cast<FtpOpenedDirectory *>(directory->getData());

	if(ftpDir->iter!=ftpDir->files.end())
	return *ftpDir->iter++;
	else
	throw ExceptionReadDirEnd();

}

void VfsFtp::changeGroup(path & source, string group)
{
	throw ExceptionNotImplemented();
}

void VfsFtp::changeOwner(path & source, string user)
{
	throw ExceptionNotImplemented();
}

void VfsFtp::chmod(path & source, unsigned int mode)
{
	string cmd="CHMOD "+FilesEngine::permissionsToOctetString(mode)+" "+source.string();
	if(!(connected && ftp.isConnected() && ftp.Site(cmd.c_str())==1))
	throw ExceptionFileSystemOther(source.string(),"chmod");
	
}

void VfsFtp::rename(path & source, path & destination)
{
	if(!(connected && ftp.isConnected() && ftp.Rename(source.string().c_str(),destination.string().c_str())==1))
	throw ExceptionFileSystemOther(source.string(),"rename");
}

void VfsFtp::removeDirectory(path & source)
{
	if(!(connected && ftp.isConnected() && ftp.Rmdir(source.string().c_str())==1))
	throw ExceptionFileSystemOther(source.string(),"removeDirectory");
	
}

void VfsFtp::removeFile(path & source)
{
	if(!(connected && ftp.isConnected() && ftp.Delete(source.string().c_str())==1))
	throw ExceptionFileSystemOther(source.string(),"removeFile");
}


void VfsFtp::init(Panel * _panel, QuestionManager * _questionManager)
{
	VfsBase::init(_panel,_questionManager);
	params=FilesEngine::parseParam(panel->getParam());	
	
	panel->setLabel(params["hostname"]+":","network_small");
	
	setHeaders(settings.getHeaders("Ftp"));
}

string VfsFtp::getType()
{
	return "remote";
}

void VfsFtp::copyFromFile(File & source, path & destination, Status * status, Cancelable * cancelable)
{
	if(!ftp.isConnected())
	throw ExceptionConnectionLost(connectionDescription,"");

	if(source.isDirectory() && !source.isSymlink())
	throw ExceptionMustRecursive();

	ftphandle* handle=ftp.RawOpen(source.getPath().string().c_str(),ftplib::fileread,  ftplib::image);
	
	int dst;
	int len=1;
	char data[BLOCK_SIZE];
	
	if(handle)
	{
	dst=open(destination.string().c_str(),O_WRONLY | O_CREAT | O_TRUNC,S_IRUSR+S_IWUSR);
		
		if(dst)
		{	
			while((len=ftp.RawRead(data,BLOCK_SIZE,handle)) > 0)
			{
					if(write(dst,data,len)!=len)
					{
					close(dst);
    					FilesEngine::translateErrnoToException(destination.string(),"write");
					}
            		
			}
		close(dst);	
		}
		else
		{
		ftp.RawClose ( handle );
    		FilesEngine::translateErrnoToException(destination.string(),"open");
		}	
				
	ftp.RawClose ( handle );
	}
	else
	throw ExceptionFileSystemOther(source.getPath().string(),"open");		
	 
}

void VfsFtp::copyToFile(File & source, path & destination, Status * status, Cancelable * cancelable)
{
	if(!ftp.isConnected())
	throw ExceptionConnectionLost(connectionDescription,"");
	
	
	if(source.isDirectory() && !source.isSymlink())
	throw ExceptionMustRecursive();
	
	
	ftphandle* handle=ftp.RawOpen(destination.string().c_str(),ftplib::filewrite,  ftplib::image);
	
	int src;
	int len=1;
	char data[BLOCK_SIZE];
	
	if(handle)
	{
		src=open(source.getPath().string().c_str(),O_RDONLY,NULL);
		
		if(src)
		{	
			while((len=read(src,data,131072)) > 0)
			{
       				if(ftp.RawWrite(data,len,handle)!=len)
				{
            			close(src);
    				throw ExceptionFileSystemOther(destination.string(),"write");
				}

			
			}
		close(src);		
		}
		else
		{
		ftp.RawClose ( handle );
		FilesEngine::translateErrnoToException(source.getPath().string(),"open");
		}
				
				
	ftp.RawClose ( handle );
	}	
	else
	throw ExceptionFileSystemOther(destination.string(),"open");	
	
}

File VfsFtp::readFile(path filePath)
{

	path p=filePath.branch_path();
			
		try
		{
		Directory directory=openDir(p);
		
			while(1)
			{
				
				File nextFile= directory.getNextFile();
				if(nextFile.getName()==filePath.leaf())
				{
				directory.close();
				return nextFile;
				}
					
			}
		}
		catch(exception e)
		{
			throw ExceptionReadFile();
		}
	

}

void VfsFtp::createDir(path p)
{
	if(!(connected && ftp.isConnected() && ftp.Mkdir(p.string().c_str())==1))
	throw ExceptionFileSystemOther(p.string(),"createDir");
}



