/***************************************************************************
 *   Copyright (C) 2007 by Mateusz Dworak aka compbatant                   *
 *   compbatant@linux.pl                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "VfsLocal.hpp"
#include "FilesEngine.hpp"
#include "Exceptions.hpp"




#include <sys/types.h>
#include <dirent.h>
#include <sys/stat.h>
#include <unistd.h>
#include <pwd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <grp.h>



VfsLocal::VfsLocal()
 : VfsBase("Local")
{
	
}


VfsLocal::~VfsLocal()
{
}



Directory VfsLocal::openDir( path& p, Cancelable *cancelable )
{
	DIR *dp = opendir (p.string().c_str());
	if (dp == NULL)
	throw ExceptionDirOpen();
	
	Directory directory(this,(void*)dp,p);
	return directory;

}

File VfsLocal::readDir( Directory *directory )
{
	
    DIR *dp=(DIR *)directory->getData();
    dirent entry;
    dirent *result;
    
    if (dp == NULL)
    {
	throw ExceptionDirOpen();
    }
    
    do
    {

  	
    	
    	int return_code = readdir_r(dp, &entry, &result);
    
	if (!(result != NULL && return_code == 0))
	{
		//closedir (dp);
		throw ExceptionReadDirEnd();
	}  
    }
    while(string(entry.d_name)=="." || (string(entry.d_name)==".." && directory->getDir()=="/"));


    	
    path fileName=path(directory->getDir()/path(entry.d_name));
    File file=readFile(fileName); 
    
    
    return readFullFile(file);  	
	
	
}


void VfsLocal::createDir( path p )
{
	if(mkdir(p.string().c_str(),S_IWUSR|S_IXUSR|S_IRUSR)==-1)
		FilesEngine::translateErrnoToException(p.string(),"createDir");
}


File VfsLocal::readFile(path fileName)
{
	File file;
	
	
	file.setName( fileName.leaf() );
	file.setParentDirectory( fileName.branch_path());
	
	
	struct stat fileStatus;
	
		if (lstat (fileName.string().c_str(), &fileStatus) != -1)
		{
			file.setSize(fileStatus.st_size);
			file.setUid(fileStatus.st_uid);
			file.setGid(fileStatus.st_gid);

			
			file.setAccessTime(fileStatus.st_atime);
			file.setModificationTime(fileStatus.st_mtime);
			file.setCreationTime(fileStatus.st_ctime);
			
			struct passwd *pwd=getpwuid(fileStatus.st_uid);
  			if(pwd)
  			{
  			file.setUser(pwd->pw_name);
  			}
		
			struct group *grp=getgrgid(fileStatus.st_gid);
  			if(grp) 
  			{
  				file.setGroup(grp->gr_name);
  			}
  			
		
			if (S_ISLNK (fileStatus.st_mode))
			{
			
				file.setSymlink(symlink(fileName).string());
				struct stat fileStatus2;
				if (stat (fileName.string().c_str(), &fileStatus2) != -1)
				{
					if (S_ISDIR (fileStatus2.st_mode))
					file.setDirectory( );
				}
				
			}			
			else if (S_ISDIR (fileStatus.st_mode))
			{
				file.setDirectory( );
			}			
			else if (S_ISCHR (fileStatus.st_mode))
			{
				file.setCharacterDevice();
			}			
			else if (S_ISBLK (fileStatus.st_mode))
			{
				file.setBlockDevice();
			}
			else if (S_ISFIFO (fileStatus.st_mode))
			{
				file.setFifo();
			}
			else if (S_ISSOCK (fileStatus.st_mode))
			{
				file.setSocket();
			}
			
			
			file.setMode(fileStatus.st_mode);

		}
		else
		{
		throw ExceptionReadFile();
		}
	

	return file;
}



void VfsLocal::copyToFile(File &source, path & destination, Status * status,  Cancelable *cancelable)
{
	copyFile(source,destination,status,cancelable);
}

void VfsLocal::copyFromFile(File &source, path & destination, Status * status,  Cancelable *cancelable)
{
	copyFile(source,destination,status,cancelable);
}

void VfsLocal::closeDir(Directory * directory)
{
	closedir ((DIR *)directory->getData());
}

void VfsLocal::copyFile(File &source, path & destination, Status * status,  Cancelable *cancelable)
{

	if(source.isDirectory() && !source.isSymlink())
	throw ExceptionMustRecursive();

	string srcFile=source.getPath().string();
	string dstFile=destination.string();
		
	
	if(status)
	status->setCurrentFileName(srcFile);
		
	struct stat fileStatus;	
		
	if(source.isBlockDevice() || source.isCharacterDevice() || source.isSocket() || source.isFifo() || source.isSymlink())
	{
		if (::stat (srcFile.c_str(), &fileStatus) == 0)
		{
			if(source.isFifo())
			{
				if(mkfifo (dstFile.c_str(), fileStatus.st_mode)!=0)
				FilesEngine::translateErrnoToException(dstFile,"mkfifo");
			}
			else if(source.isSymlink())
			{	
				path src=source.getPath();
				createSymlink(src,destination);
			}
			else
			{
				if(mknod (dstFile.c_str(), fileStatus.st_mode, fileStatus.st_rdev)!=0)
				FilesEngine::translateErrnoToException(dstFile,"mknod");
			}
		}
		else
		FilesEngine::translateErrnoToException(srcFile,"stat");
	}

		
	unsigned char buffer[4096];	
	long nread, nwritten;
	int src, dst;
	
	if ((src = open (srcFile.c_str(), O_RDONLY)) >= 0)
	{
		if (::stat (srcFile.c_str(), &fileStatus) == 0)
		{	
			if(status)
			status->setCurrentFileSize(fileStatus.st_size);
			
			if ((dst = open (dstFile.c_str(), O_WRONLY | O_CREAT | O_TRUNC, fileStatus.st_mode)) >= 0)
			{
				while (1)
				{

					if(cancelable && cancelable->isCanceledExecution())
					break;
				
				nread = fullread (src, buffer, sizeof (buffer));
				if (nread < 0)
					goto err;
				if (nread == 0)
					break;
					
				if(status)	
				status->increaseCurrentProgress(nread);
				
				nwritten = fullwrite (dst, buffer, nread);
				if (nwritten < 0)
					goto err;
				}
			err:close (dst);
			}
			else
			FilesEngine::translateErrnoToException(dstFile,"create");
		}
		else
		{
		close (src);
		FilesEngine::translateErrnoToException(srcFile,"stat");
		}
		
		close (src);
	}
	else
	FilesEngine::translateErrnoToException(srcFile,"open");
}

long VfsLocal::fullread (int fd, unsigned char *ptr, long len)
{
    long nread;
#ifdef EINTR
    do
    {
	nread = read (fd, ptr, len);
    }
    while (nread < 0 && errno == EINTR);
#else
    nread = read (fd, ptr, len);
#endif
    return nread;
}


// Write bytes
long VfsLocal::fullwrite (int fd, const unsigned char *ptr, long len)
{
    long nwritten, ntotalwritten = 0;
    while (len > 0)
    {
	nwritten = write (fd, ptr, len);
	if (nwritten < 0)
	{
#ifdef EINTR
	    if (errno == EINTR)
		continue;
#endif
	    return -1;
	}
	ntotalwritten += nwritten;
	ptr += nwritten;
	len -= nwritten;
    }
    return ntotalwritten;
}

string VfsLocal::getType()
{
	return "Local";
}



path VfsLocal::symlink(const path & file)
{
	
	char lnk[MAXPATHLEN+1];
	int len=::readlink(file.string().c_str(),lnk,MAXPATHLEN);
		if(0<=len)
		{
			lnk[len] = '\0';
			return path(string(lnk));
		}
		else
			return path(); /*TODO*/
	
}

void VfsLocal::chmod(path & source, unsigned int mode)
{
	if(::chmod(source.string().c_str(),mode)!=0)
	FilesEngine::translateErrnoToException(source.string(),"chmod");
}

void VfsLocal::removeFile(path & source)
{	
	if(::remove(source.string().c_str())!=0)
	FilesEngine::translateErrnoToException(source.string(),"remove");
}

void VfsLocal::removeDirectory(path & source)
{
	if(::remove(source.string().c_str())!=0)
	FilesEngine::translateErrnoToException(source.string(),"remove");
}

void VfsLocal::rename(path & source, path & destination)
{
	if (::rename (source.string().c_str(), destination.string().c_str()) != 0)
	FilesEngine::translateErrnoToException(source.string(),"rename");
}

void VfsLocal::changeOwner(path & source, string user)
{
	passwd *p = getpwnam( user.c_str() );
	if(p!=NULL)
	{
		if(chown (source.string().c_str(), p->pw_uid, (gid_t)-1)!=0)
		FilesEngine::translateErrnoToException(source.string(),"chown");
	}
}

vector< string > VfsLocal::getGroups()
{
	vector<string>names;
	
	struct passwd *pwde;
	while ((pwde = getpwent ()))
	{
		names.push_back (pwde->pw_name);
	}
	endpwent ();
	return names;
}

vector< string > VfsLocal::getUsers()
{
	vector<string>names;
	struct group *grpe;
	while ((grpe = getgrent ()))
	{
		names.push_back(grpe->gr_name);
	}
	endgrent ();
	return names;
}

void VfsLocal::changeGroup(path & source, string group)
{
	::group * grp = getgrnam (group.c_str ());
	if(grp!=NULL)
	{
		if(::chown (source.string().c_str(), (uid_t)-1, grp->gr_gid)!=0)
		FilesEngine::translateErrnoToException(source.string(),"chgrp");
	}
}

void VfsLocal::createSymlink(path &source,path &destination)
{
	 if(::symlink(source.string().c_str(),destination.string().c_str())!=0)
	 FilesEngine::translateErrnoToException(source.string(),"symlink");
}

void VfsLocal::init(Panel * panel, QuestionManager * questionManager)
{
	VfsBase::init(panel,questionManager);
	setHeaders(settings.getHeaders("Local"));
}


