/***************************************************************************
 *   Copyright (C) 2006 by Mateusz Dworak aka compbatant                   *
 *   m@nao.linux.pl                                                        *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "VfsSftp.hpp"
#include "Exceptions.hpp"
#include "FilesEngine.hpp"
#include "Question.hpp"

#include <sys/stat.h>
#include <fcntl.h>

#include <boost/lexical_cast.hpp>
using boost::lexical_cast;

REGISTER_FACTORABLE(VfsSftp);


VfsSftp::VfsSftp():VfsBase("remote"),connected(false),sftp(NULL),session(NULL),dirsftp(NULL)
{
}

VfsSftp::~VfsSftp()
{
	disconnect();
}

void VfsSftp::disconnect(void )
{
	if(sftp)
	sftp_free(sftp);
	sftp=NULL;
	
	if(connected)
	{
		if(session)
		{
		ssh_disconnect(session);
		ssh_finalize();
		session=NULL;
		}
	}
}


Directory VfsSftp::openDir(path &p, Cancelable *cancelable)
{
	if(!connected)
	connect();	

		if(p.string()=="")
		{
			p=path(string(sftp_canonicalize_path(sftp, "./")));
		}
		
		
		dirsftp=sftp_opendir(sftp,(char*)p.string().c_str());
	
		if(!dirsftp) 
		{
			throw ExceptionDirOpen();
		}
		else
		{
			Directory directory(this,dirsftp,p);
			return directory;
		}

}

void VfsSftp::closeDir(Directory * directory)
{
	if(sftp_dir_close((SFTP_DIR *)directory->getData())!=0)
	translateErrnoToException(directory->getDir().string(),"closeDir");
}


File VfsSftp::readDir(Directory *directory)
{
	
	SFTP_ATTRIBUTES *rfile;
	if(sftp && (rfile=sftp_readdir(sftp,(SFTP_DIR *)directory->getData()))!=NULL)
	{
		return readFullFile(readRFile(rfile,directory->getDir(),string(rfile->name)));
	}
	else
	{
		throw ExceptionReadDirEnd();
	}
}


File VfsSftp::readFile(path filePath)
{

	SFTP_ATTRIBUTES *rfile;	
	if(sftp)
	{
		rfile=sftp_lstat(sftp, (char*)filePath.string().c_str());
		if(rfile)
		{
		return readRFile(rfile,filePath.branch_path(),filePath.leaf());
		}
	}

	throw ExceptionReadFile();

}


File VfsSftp::readRFile(SFTP_ATTRIBUTES * rfile,path parentDir,string name)
{

	File file;
	file.setName(name);
	file.setSize(rfile->size);
	
	if((rfile->permissions & S_IFMT) == S_IFDIR)
	file.setDirectory();
	
	file.setParentDirectory(parentDir);	
	file.setAccessTime(rfile->atime);
	file.setModificationTime(rfile->mtime);	
	file.setMode(rfile->permissions);
	file.setUser(lexical_cast<string>(rfile->uid));
	file.setGroup(lexical_cast<string>(rfile->gid));
	sftp_attributes_free(rfile);
	return file;
	
}


void VfsSftp::createDir(path p)
{
	if(sftp)
	{
		if(sftp_mkdir(sftp, (char*)p.string().c_str(), NULL)!=0)
		translateErrnoToException(p.string(),"createDir");
	}
	else
	translateErrnoToException(p.string(),"createDir");
}

void VfsSftp::copyToFile(File &source, path & destination, Status * status, Cancelable *cancelable)
{

	if(source.isDirectory() && !source.isSymlink())
	{
	throw ExceptionMustRecursive();
	}
	
	if(!sftp)
	return;


	SFTP_FILE *dst;
	int src;
	int len=1;
    	char data[BLOCK_SIZE]; 

	if((src=open(source.getPath().string().c_str(),O_RDONLY,NULL)) >=0)
	{	
    		dst=sftp_open(sftp,(char*)destination.string().c_str(),O_WRONLY|O_CREAT,NULL);
		if(dst)
		{
			
		
			while((len=read(src,data,131072)) > 0)
			{
       				if(sftp_write(dst,data,len)!=len)
				{
				throw ExceptionWithDescription("error writing %d bytes : %s\n"+string(ssh_get_error(session)));
             			}

			
			}
		sftp_file_close(dst);		
    		}
    		else
    		{
    		close(src);
    		translateErrnoToException(destination.string(),"write");
    		}
    			
    	close(src);
    	}
    	else
    	FilesEngine::translateErrnoToException(destination.string(),"open");		
    			

}


void VfsSftp::copyFromFile(File &source, path & destination, Status * status, Cancelable *cancelable)
{

	if(source.isDirectory() && !source.isSymlink())
	{
	throw ExceptionMustRecursive();
	}
	
	SFTP_FILE *src;
	int dst;
	int len=1;
	char data[BLOCK_SIZE]; 
		
    	src=sftp_open(sftp,(char*)source.getPath().string().c_str(),O_RDONLY,NULL);
    	
		if(src)
		{
			
			if ((dst=open(destination.string().c_str(),O_WRONLY | O_CREAT | O_TRUNC,S_IRUSR+S_IWUSR)) >= 0)
			{
				while((len=sftp_read(src,data,BLOCK_SIZE)) > 0)
				{
					if(write(dst,data,len)!=len)
					{
						throw ExceptionWithDescription("error writing %d bytes : %s\n"+string(ssh_get_error(session)));
					}
				
				}
			
			close(dst);
			}
			else
			{
			sftp_file_close(src);
			FilesEngine::translateErrnoToException(destination.string(),"create");
			}
			
		sftp_file_close(src);
  		}
  		else
  		translateErrnoToException(source.getPath().string(),"open");	


}


void VfsSftp::connect(void )
{
	
	string user=params["user"];
	if(user=="")
	user=FilesEngine::getCurrentUser();
	
	string password=params["password"];;
	string hostname=params["hostname"];
	string port=params["port"];
	if(port=="")
	port="22";
		
	connectionDescription=user+"@"+hostname;
		
	SSH_OPTIONS *options;
	int auth=0;
	//char *password;
	char *banner;
	int state;
	char buf[10];
	unsigned char hash[MD5_DIGEST_LEN];
	
	options=ssh_options_new();
	ssh_options_set_username(options,(char*)user.c_str());
	ssh_options_set_host(options,(char*)hostname.c_str());    
	ssh_options_set_port(options,atoi(port.c_str()));
	
	session=ssh_new();
	ssh_set_options(session,options);
	if(ssh_connect(session))
	{	
		disconnect();
		throw ExceptionConnectionRefused(connectionDescription,"");
	}
	
	state=ssh_is_server_known(session);
	
	string textQuestion;
	
	switch(state)
	{
		case SSH_SERVER_KNOWN_OK:
		break; /* ok */
		case SSH_SERVER_KNOWN_CHANGED:
		
		ssh_get_pubkey_hash(session,hash);
		
		textQuestion="Host key for server changed : server's one is now :\n"+getHex(hash,MD5_DIGEST_LEN)+"If you really want to connect, click no";
	
		if(Question::askYesNo(questionManager,textQuestion)=="yes")
		{     
			ssh_disconnect(session);
			ssh_finalize();
			throw ExceptionNoDescription();
		}
	
		case SSH_SERVER_FOUND_OTHER:
		
		textQuestion="The host key for this server was not found but an other type of key exists.\n\
			     An attacker might change the default server key to confuse your client\n\
			     into thinking the key does not exist\n\
			     If you really want to connect, click no";
			     
		if(Question::askYesNo(questionManager,textQuestion)=="yes")
		{     
			ssh_disconnect(session);
			ssh_finalize();
			throw ExceptionNoDescription();
		}	

		case SSH_SERVER_NOT_KNOWN:
		
		ssh_get_pubkey_hash(session,hash);
		textQuestion="The server is unknown. Do you trust the host key ?\n Public key hash:\n"+getHex(hash,MD5_DIGEST_LEN);	
			
		if(Question::askYesNo(questionManager,textQuestion)=="yes")
		{
			//fprintf(stderr,"This new key will be written on disk for further usage. do you agree ?\n");	
			if(ssh_write_knownhost(session))
			{
			ssh_disconnect(session);
			throw ExceptionWithDescription("error writing key");
			}

		}
		else
		{
		ssh_disconnect(session);
		throw ExceptionNoDescription();
		}
	
	
		
		break;
		case SSH_SERVER_ERROR:		
		ssh_disconnect(session);
		ssh_finalize();
		throw ExceptionConnectionRefused(connectionDescription,"");
	}
	
	/* no ? you should :) */
	auth=ssh_userauth_autopubkey(session);
	if(auth==SSH_AUTH_ERROR)
	{
		//fprintf(stderr,"Authenticating with pubkey: %s\n",ssh_get_error(session));
		ssh_finalize();
		throw ExceptionAuthenticationFailed(connectionDescription,"");
	}
	banner=ssh_get_issue_banner(session);
	if(banner){
		printf("%s\n",banner);
		free(banner);
	}
	/*
	if(auth!=SSH_AUTH_SUCCESS){
		auth=auth_kbdint(session);
		if(auth==SSH_AUTH_ERROR){
		fprintf(stderr,"authenticating with keyb-interactive: %s\n",
			ssh_get_error(session));
		ssh_finalize();
		return -1;
		}
	}
	*/
	if(auth!=SSH_AUTH_SUCCESS)
	{
		if(ssh_userauth_password(session,NULL,(char*)password.c_str()) != SSH_AUTH_SUCCESS)
		{
		ssh_disconnect(session);
		ssh_finalize();
		throw ExceptionAuthenticationFailed(connectionDescription,"");
		}
	}
	
	//ssh_say(1,"Authentication success\n");
	
	sftp=sftp_new(session);
	
		if(sftp_init(sftp))
		{
		disconnect();
		throw ExceptionWithDescription(string("error initialising sftp :")+string(ssh_get_error(session)));
		}

}

void VfsSftp::setFileMime(vector< string > & fields, File & file)
{
	fields.push_back(file.getMime());
}

void VfsSftp::setFileSize(vector< string > & fields, File & file)
{
	fields.push_back(FilesEngine::printSize(file.getSize()));
}


string VfsSftp::getType()
{
	return "remote";
}

void VfsSftp::init(Panel * _panel,QuestionManager *_questionManager)
{
	VfsBase::init(_panel,_questionManager);
	params=FilesEngine::parseParam(panel->getParam());	
	
	panel->setLabel(params["hostname"]+":","network_small");
	
	setHeaders(settings.getHeaders("Sftp"));
	
}


void VfsSftp::chmod(path & source, unsigned int mode)
{
	SFTP_ATTRIBUTES * attr=sftp_stat (sftp,(char*)source.string().c_str ());
	if(attr) 
 	{  
	attr->permissions=mode;		
		if(sftp_setstat(sftp,(char*)source.string().c_str (),attr)!=0)
		{
			sftp_attributes_free(attr);
			translateErrnoToException(source.string(),"chmod");			
		} 	
	sftp_attributes_free(attr);
	}
	else
	{
		throw ExceptionFileSystemOther(source.string(),"chmod");
	}
}

void VfsSftp::removeFile(path & source)
{
	if(sftp)
	{
		if(sftp_rm(sftp,(char*)source.string().c_str())!=0)
		{
			translateErrnoToException(source.string(),"removeFile");
		}
	}
	else
	translateErrnoToException(source.string(),"removeFile");
}

void VfsSftp::removeDirectory(path & source)
{

	if(sftp)
	{
		if(sftp_rmdir(sftp,(char*)source.string().c_str())!=0)
		{
			translateErrnoToException(source.string(),"removeDirectory");
		}
	}
	else
	translateErrnoToException(source.string(),"removeDirectory");
}


void VfsSftp::rename(path & source, path & destination)
{

	if(sftp)
	{
		if(sftp_rename(sftp,(char*)source.string().c_str(),(char*)destination.string().c_str())!=0)
		{
			translateErrnoToException(source.string(),"rename");
		}
	}
	else
	translateErrnoToException(source.string(),"rename"); 
	
	
	
}

void VfsSftp::changeOwner(path & source, string user)
{
	throw ExceptionNotImplemented();	
}

void VfsSftp::changeGroup(path & source, string group)
{
	throw ExceptionNotImplemented();
}

void VfsSftp::translateErrnoToException(string fileName, string operation)
{
	if(!sftp)
	throw ExceptionConnectionLost(connectionDescription,operation);

	if(session)
	{
		int errorCode=ssh_get_error_code(session);
		if(errorCode==SSH_FATAL)
		{
			throw ExceptionConnectionLost(connectionDescription,operation);
		}
	}
	else
	{
		throw ExceptionConnectionLost(connectionDescription,operation);
	}
	
	throw ExceptionWithDescription("error");
}




string VfsSftp::getHex(unsigned char * what, int len)
{
	char tab[4];

	string hexstring;

	 for(int i=0;i<len-1;++i)
	 {	 	 
       	 	sprintf(tab,"%.2hhx:",what[i]);
       	 	hexstring=hexstring+tab;
        	if((i+1) % 16 ==0)
            	hexstring=hexstring+"\n";
    	 }
    	//sprintf(tab,"%.2hhx\n",what[i]);

    	return hexstring;
}

